from typing import List

import pytest
from testapp.models import Bookmark, Message, Profile, Tagged, Thread, User

from pydantic import ConfigDict
from djantic import ModelSchema


@pytest.mark.django_db
def test_get_instance():
    """
    Test retrieving an existing Django object to populate the schema model.
    """

    user = User.objects.create(
        first_name="Jordan", last_name="Eremieff", email="jordan@eremieff.com"
    )

    class UserSchema(ModelSchema):
        model_config = ConfigDict(model=User, include=["id", "first_name"])

    assert UserSchema.from_django(user).model_dump() == {
        "first_name": "Jordan",
        "id": 1,
    }


@pytest.mark.django_db
def test_get_instance_with_generic_foreign_key():

    bookmark = Bookmark.objects.create(url="https://www.djangoproject.com/")
    Tagged.objects.create(content_object=bookmark, slug="django")

    class TaggedSchema(ModelSchema):
        model_config = ConfigDict(model=Tagged)

    class BookmarkWithTaggedSchema(ModelSchema):

        tags: List[TaggedSchema]
        model_config = ConfigDict(model=Bookmark)

    bookmark_with_tagged_schema = BookmarkWithTaggedSchema.from_django(bookmark)

    assert bookmark_with_tagged_schema.model_dump() == {
        "id": 1,
        "tags": [
            {
                "content_object": 1,
                "content_type": 20,
                "id": 1,
                "object_id": 1,
                "slug": "django",
            }
        ],
        "url": "https://www.djangoproject.com/",
    }


@pytest.mark.django_db
def test_get_queryset_with_reverse_one_to_one():
    """
    Test retrieving a Django queryset with reverse one-to-one relationships.
    """
    user_data = [
        {"first_name": "Jordan", "email": "jordan@eremieff.com"},
        {"first_name": "Sara", "email": "sara@example.com"},
    ]
    for kwargs in user_data:
        user = User.objects.create(**kwargs)
        Profile.objects.create(user=user, location="Australia")

    class UserSchema(ModelSchema):
        model_config = ConfigDict(
            model=User,
            include=["id", "email", "first_name", "profile"],
        )

    users = User.objects.all()
    user_schema_qs = UserSchema.from_django(users, many=True)
    assert user_schema_qs == [
        {
            "email": "jordan@eremieff.com",
            "first_name": "Jordan",
            "id": 1,
            "profile": 1,
        },
        {"email": "sara@example.com", "first_name": "Sara", "id": 2, "profile": 2},
    ]

    # Test when using a declared sub-model
    class ProfileSchema(ModelSchema):
        model_config = ConfigDict(model=Profile, include=["id", "location"])

    class UserWithProfileSchema(ModelSchema):
        profile: ProfileSchema
        model_config = ConfigDict(
            model=User, exclude=["created_at", "updated_at", "last_name"]
        )

    users = User.objects.all()

    user_with_profile_schema_qs = UserWithProfileSchema.from_django(users, many=True)
    assert user_with_profile_schema_qs == [
        {
            "email": "jordan@eremieff.com",
            "first_name": "Jordan",
            "id": 1,
            "profile": {"id": 1, "location": "Australia"},
        },
        {
            "email": "sara@example.com",
            "first_name": "Sara",
            "id": 2,
            "profile": {"id": 2, "location": "Australia"},
        },
    ]


@pytest.mark.django_db
def test_get_queryset_with_foreign_key():
    """
    Test retrieving a Django queryset with foreign-key relationships.
    """

    thread = Thread.objects.create(title="My thread topic")
    thread2 = Thread.objects.create(title="Another topic")
    for content in ("I agree.", "I disagree!", "lol"):
        message_one = Message.objects.create(content=content, thread=thread)
        Message.objects.create(content=content, thread=thread2)

    class MessageSchema(ModelSchema):
        model_config = ConfigDict(model=Message, exclude=["created_at"])

    schema = MessageSchema.from_django(message_one)
    assert schema.model_dump() == {"id": 5, "content": "lol", "thread": 1}

    class ThreadSchema(ModelSchema):
        model_config = ConfigDict(model=Thread, exclude=["created_at"])

    class MessageWithThreadSchema(ModelSchema):
        thread: ThreadSchema
        model_config = ConfigDict(model=Message, exclude=["created_at"])

    schema = MessageWithThreadSchema.from_django(message_one)
    assert schema.model_dump() == {
        "id": 5,
        "content": "lol",
        "thread": {
            "messages": [{"id": 1}, {"id": 3}, {"id": 5}],
            "id": 1,
            "title": "My thread topic",
        },
    }


@pytest.mark.django_db
def test_get_queryset_with_reverse_foreign_key():
    """
    Test retrieving a Django queryset with reverse foreign-key relationships.
    """

    thread = Thread.objects.create(title="My thread topic")
    thread2 = Thread.objects.create(title="Another topic")
    for content in ("I agree.", "I disagree!", "lol"):
        Message.objects.create(content=content, thread=thread)
        Message.objects.create(content=content, thread=thread2)

    threads = Thread.objects.all()

    class MessageSchema(ModelSchema):
        model_config = ConfigDict(model=Message, include=["id", "content"])

    class ThreadSchema(ModelSchema):
        model_config = ConfigDict(model=Thread)

    thread_schema_qs = ThreadSchema.from_django(threads, many=True)
    thread_schemas = [t.model_dump() for t in thread_schema_qs]
    assert thread_schemas == [
        {
            "messages": [{"id": 2}, {"id": 4}, {"id": 6}],
            "id": 2,
            "title": "Another topic",
        },
        {
            "messages": [{"id": 1}, {"id": 3}, {"id": 5}],
            "id": 1,
            "title": "My thread topic",
        },
    ]

    # Test when using a declared sub-model
    class ThreadWithMessageListSchema(ModelSchema):
        messages: List[MessageSchema]
        model_config = ConfigDict(model=Thread, exclude=["created_at", "updated_at"])

    thread_with_message_list_schema_qs = ThreadWithMessageListSchema.from_django(
        threads, many=True
    )

    assert thread_with_message_list_schema_qs == [
        {
            "messages": [
                {"id": 2, "content": "I agree."},
                {"id": 4, "content": "I disagree!"},
                {"id": 6, "content": "lol"},
            ],
            "id": 2,
            "title": "Another topic",
        },
        {
            "messages": [
                {"id": 1, "content": "I agree."},
                {"id": 3, "content": "I disagree!"},
                {"id": 5, "content": "lol"},
            ],
            "id": 1,
            "title": "My thread topic",
        },
    ]


@pytest.mark.django_db
def test_get_queryset_with_generic_foreign_key():

    bookmark = Bookmark.objects.create(url="https://github.com")
    bookmark.tags.create(slug="tag-1")
    bookmark.tags.create(slug="tag-2")

    class TaggedSchema(ModelSchema):
        model_config = ConfigDict(model=Tagged)

    class BookmarkSchema(ModelSchema):
        model_config = ConfigDict(model=Bookmark)

    schema = BookmarkSchema.from_django(bookmark)
    schema.model_dump() == {
        "id": 1,
        "url": "https://github.com",
        "tags": [{"id": 1}, {"id": 2}],
    }



from typing import Optional

import pytest
from pydantic import ConfigDict, ValidationError, ValidationInfo, field_validator
from testapp.models import (
    Configuration,
    Listing,
    NullableChar,
    NullableFK,
    Preference,
    Record,
    Searchable,
    User,
)

from djantic import ModelSchema


@pytest.mark.django_db
def test_schema_without_include_and_exclude():
    """
    Using a schema without include and exclude populates userschema correctly
    Optional foreign key handled gracefully.
    """

    user = User.objects.create(
        first_name="Jordan", last_name="Eremieff", email="jordan@eremieff.com"
    )

    class UserSchema(ModelSchema):
        model_config = ConfigDict(model=User)

    dumped = UserSchema.from_django(user).model_dump()
    # These values are here, but a hassle to use freeze gun.
    dumped.pop("updated_at")
    dumped.pop("created_at")

    assert dumped == {
        "first_name": "Jordan",
        "id": 1,
        "email": "jordan@eremieff.com",
        "profile": None,
        "last_name": "Eremieff",
    }


@pytest.mark.django_db
def test_context_for_field():

    def get_context():
        return {'check_title': lambda x: x.istitle()}

    class UserSchema(ModelSchema):
        model_config = ConfigDict(
            model=User,
            revalidate_instances='always'
        )

        @field_validator('first_name', mode="before", check_fields=False)
        @classmethod
        def validate_first_name(cls, v: str, info: ValidationInfo):
            if not info.context:
                return v

            check_title = info.context.get('check_title')
            if check_title and not check_title(v):
                raise ValueError('First name needs to be a title')
            return v

    user = User.objects.create(first_name="hello", email="a@a.com")
    with pytest.raises(ValidationError):
        UserSchema.from_django(user, context=get_context())


@pytest.mark.django_db
def test_unhandled_field_type():
    class SearchableSchema(ModelSchema):
        model_config = ConfigDict(model=Searchable)

    assert SearchableSchema.model_json_schema() == {
        "description": "Searchable(id, title, search_vector)",
        "properties": {
            "id": {
                "anyOf": [{"type": "integer"}, {"type": "null"}],
                "default": None,
                "description": "id",
                "title": "Id",
            },
            "title": {
                "description": "title",
                "maxLength": 255,
                "title": "Title",
                "type": "string",
            },
            "search_vector": {
                "anyOf": [{"type": "string"}, {"type": "null"}],
                "default": None,
                "description": "search_vector",
                "title": "Search Vector",
            },
        },
        "required": ["title"],
        "title": "SearchableSchema",
        "type": "object",
    }
    searchable = Searchable.objects.create(title="My content")
    assert SearchableSchema.from_django(searchable).model_dump() == {
        "id": 1,
        "title": "My content",
        "search_vector": None,
    }


@pytest.mark.django_db
def test_custom_field():
    """
    Test a model using custom field subclasses.
    """

    class RecordSchema(ModelSchema):
        model_config = ConfigDict(model=Record, include=["id", "title", "items"])

    assert RecordSchema.model_json_schema() == {
        "description": "A generic record model.",
        "properties": {
            "id": {
                "anyOf": [{"type": "integer"}, {"type": "null"}],
                "default": None,
                "description": "id",
                "title": "Id",
            },
            "title": {
                "description": "title",
                "maxLength": 20,
                "title": "Title",
                "type": "string",
            },
            "items": {
                "anyOf": [
                    {
                        "contentMediaType": "application/json",
                        "contentSchema": {},
                        "type": "string",
                    },
                    {"type": "object"},
                    {"items": {}, "type": "array"},
                ],
                "description": "items",
                "title": "Items",
            },
        },
        "required": ["title"],
        "title": "RecordSchema",
        "type": "object",
    }


@pytest.mark.django_db
def test_postgres_json_field():
    """
    Test generating a model_json_schema for multiple Postgres JSON fields.
    """

    class ConfigurationSchema(ModelSchema):
        model_config = ConfigDict(
            model=Configuration, include=["permissions", "changelog", "metadata"]
        )

    assert ConfigurationSchema.model_json_schema() == {
        "description": "A configuration container.",
        "properties": {
            "permissions": {
                "anyOf": [
                    {
                        "contentMediaType": "application/json",
                        "contentSchema": {},
                        "type": "string",
                    },
                    {"type": "object"},
                    {"items": {}, "type": "array"},
                ],
                "description": "permissions",
                "title": "Permissions",
            },
            "changelog": {
                "anyOf": [
                    {
                        "contentMediaType": "application/json",
                        "contentSchema": {},
                        "type": "string",
                    },
                    {"type": "object"},
                    {"items": {}, "type": "array"},
                ],
                "description": "changelog",
                "title": "Changelog",
            },
            "metadata": {
                "anyOf": [
                    {
                        "contentMediaType": "application/json",
                        "contentSchema": {},
                        "type": "string",
                    },
                    {"type": "object"},
                    {"items": {}, "type": "array"},
                    {"type": "null"},
                ],
                "default": None,
                "description": "metadata",
                "title": "Metadata",
            },
        },
        "title": "ConfigurationSchema",
        "type": "object",
    }


@pytest.mark.django_db
def test_lazy_choice_field():
    """
    Test generating a dynamic enum choice field.
    """

    class RecordSchema(ModelSchema):
        model_config = ConfigDict(
            model=Record, include=["record_type", "record_status"]
        )

    assert RecordSchema.model_json_schema() == {
        "$defs": {
            "RecordSchemaRecordStatusEnum": {
                "enum": [0, 1, 2],
                "title": "RecordSchemaRecordStatusEnum",
                "type": "integer",
            },
            "RecordSchemaRecordTypeEnum": {
                "enum": ["NEW", "OLD"],
                "title": "RecordSchemaRecordTypeEnum",
                "type": "string",
            },
        },
        "description": "A generic record model.",
        "properties": {
            "record_type": {
                "$ref": "#/$defs/RecordSchemaRecordTypeEnum",
                "default": "NEW",
                "description": "record_type",
                "title": "Record Type",
            },
            "record_status": {
                "$ref": "#/$defs/RecordSchemaRecordStatusEnum",
                "default": 0,
                "description": "record_status",
                "title": "Record Status",
            },
        },
        "title": "RecordSchema",
        "type": "object",
    }


@pytest.mark.django_db
def test_enum_choices():
    class PreferenceSchema(ModelSchema):
        model_config = ConfigDict(model=Preference, use_enum_values=True)

    assert PreferenceSchema.model_json_schema() == {
        "$defs": {
            "PreferenceSchemaPreferredFoodEnum": {
                "enum": ["ba", "ap"],
                "title": "PreferenceSchemaPreferredFoodEnum",
                "type": "string",
            },
            "PreferenceSchemaPreferredGroupEnum": {
                "enum": [1, 2],
                "title": "PreferenceSchemaPreferredGroupEnum",
                "type": "integer",
            },
            "PreferenceSchemaPreferredMusicianEnum": {
                "enum": ["tom_jobim", "sinatra", ""],
                "title": "PreferenceSchemaPreferredMusicianEnum",
                "type": "string",
            },
            "PreferenceSchemaPreferredSportEnum": {
                "enum": ["football", "basketball", ""],
                "title": "PreferenceSchemaPreferredSportEnum",
                "type": "string",
            },
        },
        "description": "Preference(id, name, preferred_food, preferred_group, preferred_sport, preferred_musician)",
        "properties": {
            "id": {
                "anyOf": [{"type": "integer"}, {"type": "null"}],
                "default": None,
                "description": "id",
                "title": "Id",
            },
            "name": {
                "description": "name",
                "maxLength": 128,
                "title": "Name",
                "type": "string",
            },
            "preferred_food": {
                "$ref": "#/$defs/PreferenceSchemaPreferredFoodEnum",
                "default": "ba",
                "description": "preferred_food",
                "title": "Preferred Food",
            },
            "preferred_group": {
                "$ref": "#/$defs/PreferenceSchemaPreferredGroupEnum",
                "default": 1,
                "description": "preferred_group",
                "title": "Preferred Group",
            },
            "preferred_sport": {
                "anyOf": [
                    {"$ref": "#/$defs/PreferenceSchemaPreferredSportEnum"},
                    {"type": "null"},
                ],
                "default": None,
                "description": "preferred_sport",
                "title": "Preferred Sport",
            },
            "preferred_musician": {
                "anyOf": [
                    {"$ref": "#/$defs/PreferenceSchemaPreferredMusicianEnum"},
                    {"type": "null"},
                ],
                "default": "",
                "description": "preferred_musician",
                "title": "Preferred Musician",
            },
        },
        "required": ["name"],
        "title": "PreferenceSchema",
        "type": "object",
    }

    preference = Preference.objects.create(
        name="Jordan",
        preferred_sport="",
        preferred_musician=None,
    )
    assert PreferenceSchema.from_django(preference).model_dump() == {
        "id": 1,
        "name": "Jordan",
        # default choice dumped
        "preferred_food": "ba",
        # default choice dumped
        "preferred_group": 1,
        "preferred_sport": "",
        "preferred_musician": None,
    }


@pytest.mark.skip(reason="Enums seems unique, but need other approach to verify")
@pytest.mark.django_db
def test_enum_choices_generates_unique_enums():
    class PreferenceSchema(ModelSchema):
        model_config = ConfigDict(model=Preference, use_enum_values=True)

    class PreferenceSchema2(ModelSchema):
        model_config = ConfigDict(model=Preference, use_enum_values=True)

    assert str(PreferenceSchema2.model_fields["preferred_food"].type_) != str(
        PreferenceSchema.model_fields["preferred_food"].type_
    )


@pytest.mark.django_db
def test_listing():
    """
    Testing the following properties for a field.
    - nullable foreign key
    - array field
    """

    class ListingSchema(ModelSchema):
        model_config = ConfigDict(model=Listing, use_enum_values=True)

    assert ListingSchema.model_json_schema() == {
        "description": "Listing(id, items, content_type)",
        "properties": {
            "id": {
                "anyOf": [{"type": "integer"}, {"type": "null"}],
                "default": None,
                "description": "id",
                "title": "Id",
            },
            "items": {
                "description": "items",
                "items": {"type": "string"},
                "title": "Items",
                "type": "array",
            },
            "content_type": {
                "anyOf": [{"type": "integer"}, {"type": "null"}],
                "default": None,
                "description": "id",
                "title": "Content Type",
            },
        },
        "required": ["items"],
        "title": "ListingSchema",
        "type": "object",
    }

    preference = Listing(items=["a", "b"])
    assert ListingSchema.from_django(preference).dict() == {
        "content_type": None,
        "id": None,
        "items": ["a", "b"],
    }


@pytest.mark.django_db
def test_nullable_fk():
    class NullableCharSchema(ModelSchema):
        model_config = ConfigDict(model=NullableChar, include='value')

    class NullableFKSchema(ModelSchema):
        nullable_char: Optional[NullableCharSchema] = None
        model_config = ConfigDict(model=NullableFK, include='nullable_char')

    nullable_char = NullableChar(value="test")
    nullable_char.save()
    model = NullableFK(nullable_char=nullable_char)
    assert NullableFKSchema.from_django(model).dict() == {
        "nullable_char": {
            "value": "test"
        }
    }

    model2 = NullableFK(nullable_char=None)
    assert NullableFKSchema.from_django(model2).dict() == {
        "nullable_char": None
    }



from decimal import Decimal
from typing import List, Optional

import pytest
from pydantic import validator
from testapp.order import (
    Order,
    OrderItem,
    OrderItemDetail,
    OrderUser,
    OrderUserFactory,
    OrderUserProfile,
)

from pydantic import ConfigDict
from djantic import ModelSchema


@pytest.mark.django_db
def test_multiple_level_relations():
    class OrderItemDetailSchema(ModelSchema):
        model_config = ConfigDict(model=OrderItemDetail)

    class OrderItemSchema(ModelSchema):
        details: List[OrderItemDetailSchema]
        model_config = ConfigDict(model=OrderItem)

    class OrderSchema(ModelSchema):
        items: List[OrderItemSchema]
        model_config = ConfigDict(model=Order)

    class OrderUserProfileSchema(ModelSchema):
        model_config = ConfigDict(model=OrderUserProfile)

    class OrderUserSchema(ModelSchema):
        orders: List[OrderSchema]
        profile: OrderUserProfileSchema
        user_cache: Optional[dict]
        model_config = ConfigDict(
            model=OrderUser,
            include=(
                "id",
                "first_name",
                "last_name",
                "email",
                "profile",
                "orders",
                "user_cache",
            ),
        )

        @validator("user_cache", pre=True, always=True)
        def get_user_cache(cls, _):
            return {"has_order": True}

    user = OrderUserFactory.create()

    assert OrderUserSchema.from_django(user).dict() == {
        "profile": {"id": 1, "address": "", "user": 1},
        "orders": [
            {
                "items": [
                    {
                        "details": [
                            {
                                "id": 1,
                                "name": "",
                                "value": 0,
                                "quantity": 0,
                                "order_item": 1,
                            },
                            {
                                "id": 2,
                                "name": "",
                                "value": 0,
                                "quantity": 0,
                                "order_item": 1,
                            },
                        ],
                        "id": 1,
                        "name": "",
                        "price": Decimal("0.00000"),
                        "quantity": 0,
                        "order": 1,
                    },
                    {
                        "details": [
                            {
                                "id": 3,
                                "name": "",
                                "value": 0,
                                "quantity": 0,
                                "order_item": 2,
                            },
                            {
                                "id": 4,
                                "name": "",
                                "value": 0,
                                "quantity": 0,
                                "order_item": 2,
                            },
                        ],
                        "id": 2,
                        "name": "",
                        "price": Decimal("0.00000"),
                        "quantity": 0,
                        "order": 1,
                    },
                ],
                "id": 1,
                "total_price": Decimal("0.00000"),
                "shipping_address": "",
                "user": 1,
            },
            {
                "items": [
                    {
                        "details": [
                            {
                                "id": 5,
                                "name": "",
                                "value": 0,
                                "quantity": 0,
                                "order_item": 3,
                            },
                            {
                                "id": 6,
                                "name": "",
                                "value": 0,
                                "quantity": 0,
                                "order_item": 3,
                            },
                        ],
                        "id": 3,
                        "name": "",
                        "price": Decimal("0.00000"),
                        "quantity": 0,
                        "order": 2,
                    },
                    {
                        "details": [
                            {
                                "id": 7,
                                "name": "",
                                "value": 0,
                                "quantity": 0,
                                "order_item": 4,
                            },
                            {
                                "id": 8,
                                "name": "",
                                "value": 0,
                                "quantity": 0,
                                "order_item": 4,
                            },
                        ],
                        "id": 4,
                        "name": "",
                        "price": Decimal("0.00000"),
                        "quantity": 0,
                        "order": 2,
                    },
                ],
                "id": 2,
                "total_price": Decimal("0.00000"),
                "shipping_address": "",
                "user": 1,
            },
        ],
        "id": 1,
        "first_name": "",
        "last_name": None,
        "email": "",
        "user_cache": {"has_order": True},
    }

    assert OrderUserSchema.model_json_schema() == {
        "$defs": {
            "OrderItemDetailSchema": {
                "description": "OrderItemDetail(id, name, value, quantity, order_item)",
                "properties": {
                    "id": {
                        "anyOf": [{"type": "integer"}, {"type": "null"}],
                        "default": None,
                        "description": "id",
                        "title": "Id",
                    },
                    "name": {
                        "description": "name",
                        "maxLength": 30,
                        "title": "Name",
                        "type": "string",
                    },
                    "value": {
                        "default": 0,
                        "description": "value",
                        "title": "Value",
                        "type": "integer",
                    },
                    "quantity": {
                        "default": 0,
                        "description": "quantity",
                        "title": "Quantity",
                        "type": "integer",
                    },
                    "order_item": {
                        "description": "id",
                        "title": "Order Item",
                        "type": "integer",
                    },
                },
                "required": ["name", "order_item"],
                "title": "OrderItemDetailSchema",
                "type": "object",
            },
            "OrderItemSchema": {
                "description": "OrderItem(id, name, price, quantity, order)",
                "properties": {
                    "details": {
                        "items": {"$ref": "#/$defs/OrderItemDetailSchema"},
                        "title": "Details",
                        "type": "array",
                    },
                    "id": {
                        "anyOf": [{"type": "integer"}, {"type": "null"}],
                        "default": None,
                        "description": "id",
                        "title": "Id",
                    },
                    "name": {
                        "description": "name",
                        "maxLength": 30,
                        "title": "Name",
                        "type": "string",
                    },
                    "price": {
                        "anyOf": [{"type": "number"}, {"type": "string"}],
                        "default": 0,
                        "description": "price",
                        "title": "Price",
                    },
                    "quantity": {
                        "default": 0,
                        "description": "quantity",
                        "title": "Quantity",
                        "type": "integer",
                    },
                    "order": {"description": "id", "title": "Order", "type": "integer"},
                },
                "required": ["details", "name", "order"],
                "title": "OrderItemSchema",
                "type": "object",
            },
            "OrderSchema": {
                "description": "Order(id, total_price, shipping_address, user)",
                "properties": {
                    "items": {
                        "items": {"$ref": "#/$defs/OrderItemSchema"},
                        "title": "Items",
                        "type": "array",
                    },
                    "id": {
                        "anyOf": [{"type": "integer"}, {"type": "null"}],
                        "default": None,
                        "description": "id",
                        "title": "Id",
                    },
                    "total_price": {
                        "anyOf": [{"type": "number"}, {"type": "string"}],
                        "default": 0,
                        "description": "total_price",
                        "title": "Total Price",
                    },
                    "shipping_address": {
                        "description": "shipping_address",
                        "maxLength": 255,
                        "title": "Shipping Address",
                        "type": "string",
                    },
                    "user": {"description": "id", "title": "User", "type": "integer"},
                },
                "required": ["items", "shipping_address", "user"],
                "title": "OrderSchema",
                "type": "object",
            },
            "OrderUserProfileSchema": {
                "description": "OrderUserProfile(id, address, user)",
                "properties": {
                    "id": {
                        "anyOf": [{"type": "integer"}, {"type": "null"}],
                        "default": None,
                        "description": "id",
                        "title": "Id",
                    },
                    "address": {
                        "description": "address",
                        "maxLength": 255,
                        "title": "Address",
                        "type": "string",
                    },
                    "user": {"description": "id", "title": "User", "type": "integer"},
                },
                "required": ["address", "user"],
                "title": "OrderUserProfileSchema",
                "type": "object",
            },
        },
        "description": "OrderUser(id, first_name, last_name, email)",
        "properties": {
            "profile": {"$ref": "#/$defs/OrderUserProfileSchema"},
            "orders": {
                "items": {"$ref": "#/$defs/OrderSchema"},
                "title": "Orders",
                "type": "array",
            },
            "id": {
                "anyOf": [{"type": "integer"}, {"type": "null"}],
                "default": None,
                "description": "id",
                "title": "Id",
            },
            "first_name": {
                "description": "first_name",
                "maxLength": 50,
                "title": "First Name",
                "type": "string",
            },
            "last_name": {
                "anyOf": [{"maxLength": 50, "type": "string"}, {"type": "null"}],
                "default": None,
                "description": "last_name",
                "title": "Last Name",
            },
            "email": {
                "description": "email",
                "maxLength": 254,
                "title": "Email",
                "type": "string",
            },
            "user_cache": {
                "anyOf": [{"type": "object"}, {"type": "null"}],
                "default": None,
                "title": "User Cache",
            },
        },
        "required": ["profile", "orders", "first_name", "email"],
        "title": "OrderUserSchema",
        "type": "object",
    }



import pytest
from pydantic.errors import PydanticUserError
from testapp.models import User

from pydantic import ConfigDict
from djantic import ModelSchema


@pytest.mark.django_db
def test_model_config_contains_valid_model():
    error_msg = r"(Is `model_config\[\"model\"\]` a valid Django model class?)"
    with pytest.raises(PydanticUserError, match=error_msg):

        class InvalidModelErrorSchema2(ModelSchema):
            model_config = ConfigDict(model="Ok")


@pytest.mark.django_db
def test_include_and_exclude_mutually_exclusive():
    error_msg = "Only one of 'include' or 'exclude' should be set in configuration."
    with pytest.raises(PydanticUserError, match=error_msg):

        class IncludeExcludeErrorSchema(ModelSchema):
            model_config = ConfigDict(
                model=User, include=["id"], exclude=["first_name"]
            )


@pytest.mark.django_db
def test_get_field_names():
    """
    Test retrieving the field names for a model.
    """

    class UserSchema(ModelSchema):
        model_config = ConfigDict(model=User, include=["id"])

    assert UserSchema.get_field_names() == ["id"]

    class UserSchema(ModelSchema):
        model_config = ConfigDict(model=User, exclude=["id"])

    assert UserSchema.get_field_names() == [
        "profile",
        "first_name",
        "last_name",
        "email",
        "created_at",
        "updated_at",
    ]



import datetime
from typing import Dict, List, Optional

import pytest
from pydantic import Field
from testapp.models import (
    Article,
    Bookmark,
    Case,
    Expert,
    Item,
    Message,
    Profile,
    Publication,
    Tagged,
    Thread,
    User,
)

from pydantic import ConfigDict
from djantic import ModelSchema


@pytest.mark.django_db
def test_m2m():
    """
    Test forward m2m relationships.
    """

    class ArticleSchema(ModelSchema):
        model_config = ConfigDict(model=Article)

    assert ArticleSchema.model_json_schema() == {
        "description": "A news article.",
        "properties": {
            "id": {
                "anyOf": [{"type": "integer"}, {"type": "null"}],
                "default": None,
                "description": "id",
                "title": "Id",
            },
            "headline": {
                "description": "headline",
                "maxLength": 100,
                "title": "Headline",
                "type": "string",
            },
            "pub_date": {
                "description": "pub_date",
                "format": "date",
                "title": "Pub Date",
                "type": "string",
            },
            "publications": {
                "description": "id",
                "items": {
                    "additionalProperties": {"type": "integer"},
                    "type": "object",
                },
                "title": "Publications",
                "type": "array",
            },
        },
        "required": ["headline", "pub_date", "publications"],
        "title": "ArticleSchema",
        "type": "object",
    }

    class PublicationSchema(ModelSchema):
        class Config:
            model = Publication

    class ArticleWithPublicationListSchema(ModelSchema):
        publications: List[PublicationSchema]
        model_config = ConfigDict(model=Article)

    assert ArticleWithPublicationListSchema.model_json_schema() == {
        "$defs": {
            "PublicationSchema": {
                "description": "A news publication.",
                "properties": {
                    "article_set": {
                        "default": None,
                        "description": "id",
                        "items": {
                            "additionalProperties": {"type": "integer"},
                            "type": "object",
                        },
                        "title": "Article Set",
                        "type": "array",
                    },
                    "id": {
                        "anyOf": [{"type": "integer"}, {"type": "null"}],
                        "default": None,
                        "description": "id",
                        "title": "Id",
                    },
                    "title": {
                        "description": "title",
                        "maxLength": 30,
                        "title": "Title",
                        "type": "string",
                    },
                },
                "required": ["title"],
                "title": "PublicationSchema",
                "type": "object",
            }
        },
        "description": "A news article.",
        "properties": {
            "id": {
                "anyOf": [{"type": "integer"}, {"type": "null"}],
                "default": None,
                "description": "id",
                "title": "Id",
            },
            "headline": {
                "description": "headline",
                "maxLength": 100,
                "title": "Headline",
                "type": "string",
            },
            "pub_date": {
                "description": "pub_date",
                "format": "date",
                "title": "Pub Date",
                "type": "string",
            },
            "publications": {
                "items": {"$ref": "#/$defs/PublicationSchema"},
                "title": "Publications",
                "type": "array",
            },
        },
        "required": ["headline", "pub_date", "publications"],
        "title": "ArticleWithPublicationListSchema",
        "type": "object",
    }

    article = Article.objects.create(
        headline="My Headline", pub_date=datetime.date(2021, 3, 20)
    )
    publication = Publication.objects.create(title="My Publication")
    article.publications.add(publication)

    schema = ArticleWithPublicationListSchema.from_django(article)
    assert schema.dict() == {
        "id": 1,
        "headline": "My Headline",
        "pub_date": datetime.date(2021, 3, 20),
        "publications": [
            {"article_set": [{"id": 1}], "id": 1, "title": "My Publication"}
        ],
    }


@pytest.mark.django_db
def test_foreign_key():
    """
    Test forward foreign-key relationships.
    """

    class ThreadSchema(ModelSchema):
        model_config = ConfigDict(model=Thread)

    class MessageSchema(ModelSchema):
        model_config = ConfigDict(model=Message)

    assert MessageSchema.model_json_schema() == {
        "description": "A message posted in a thread.",
        "properties": {
            "id": {
                "anyOf": [{"type": "integer"}, {"type": "null"}],
                "default": None,
                "description": "id",
                "title": "Id",
            },
            "content": {"description": "content", "title": "Content", "type": "string"},
            "created_at": {
                "description": "created_at",
                "format": "date-time",
                "title": "Created At",
                "type": "string",
            },
            "thread": {"description": "id", "title": "Thread", "type": "integer"},
        },
        "required": ["content", "created_at", "thread"],
        "title": "MessageSchema",
        "type": "object",
    }

    class MessageWithThreadSchema(ModelSchema):
        thread: ThreadSchema
        model_config = ConfigDict(model=Message)

    assert MessageWithThreadSchema.model_json_schema() == {
        "$defs": {
            "ThreadSchema": {
                "description": "A thread of messages.",
                "properties": {
                    "messages": {
                        "default": None,
                        "description": "id",
                        "items": {
                            "additionalProperties": {"type": "integer"},
                            "type": "object",
                        },
                        "title": "Messages",
                        "type": "array",
                    },
                    "id": {
                        "anyOf": [{"type": "integer"}, {"type": "null"}],
                        "default": None,
                        "description": "id",
                        "title": "Id",
                    },
                    "title": {
                        "description": "title",
                        "maxLength": 30,
                        "title": "Title",
                        "type": "string",
                    },
                },
                "required": ["title"],
                "title": "ThreadSchema",
                "type": "object",
            }
        },
        "description": "A message posted in a thread.",
        "properties": {
            "id": {
                "anyOf": [{"type": "integer"}, {"type": "null"}],
                "default": None,
                "description": "id",
                "title": "Id",
            },
            "content": {"description": "content", "title": "Content", "type": "string"},
            "created_at": {
                "description": "created_at",
                "format": "date-time",
                "title": "Created At",
                "type": "string",
            },
            "thread": {"$ref": "#/$defs/ThreadSchema"},
        },
        "required": ["content", "created_at", "thread"],
        "title": "MessageWithThreadSchema",
        "type": "object",
    }

    class ThreadWithMessageListSchema(ModelSchema):
        messages: List[MessageSchema]
        model_config = ConfigDict(model=Thread)

    assert ThreadWithMessageListSchema.model_json_schema() == {
        "$defs": {
            "MessageSchema": {
                "description": "A message posted in a thread.",
                "properties": {
                    "id": {
                        "anyOf": [{"type": "integer"}, {"type": "null"}],
                        "default": None,
                        "description": "id",
                        "title": "Id",
                    },
                    "content": {
                        "description": "content",
                        "title": "Content",
                        "type": "string",
                    },
                    "created_at": {
                        "description": "created_at",
                        "format": "date-time",
                        "title": "Created At",
                        "type": "string",
                    },
                    "thread": {
                        "description": "id",
                        "title": "Thread",
                        "type": "integer",
                    },
                },
                "required": ["content", "created_at", "thread"],
                "title": "MessageSchema",
                "type": "object",
            }
        },
        "description": "A thread of messages.",
        "properties": {
            "messages": {
                "items": {"$ref": "#/$defs/MessageSchema"},
                "title": "Messages",
                "type": "array",
            },
            "id": {
                "anyOf": [{"type": "integer"}, {"type": "null"}],
                "default": None,
                "description": "id",
                "title": "Id",
            },
            "title": {
                "description": "title",
                "maxLength": 30,
                "title": "Title",
                "type": "string",
            },
        },
        "required": ["messages", "title"],
        "title": "ThreadWithMessageListSchema",
        "type": "object",
    }


@pytest.mark.django_db
def test_one_to_one():
    """
    Test forward one-to-one relationships.
    """

    class UserSchema(ModelSchema):
        model_config = ConfigDict(model=User)

    class ProfileSchema(ModelSchema):
        model_config = ConfigDict(model=Profile)

    assert ProfileSchema.model_json_schema() == {
        "description": "A user's profile.",
        "properties": {
            "id": {
                "anyOf": [{"type": "integer"}, {"type": "null"}],
                "default": None,
                "description": "id",
                "title": "Id",
            },
            "user": {"description": "id", "title": "User", "type": "integer"},
            "website": {
                "default": "",
                "description": "website",
                "maxLength": 200,
                "title": "Website",
                "type": "string",
            },
            "location": {
                "default": "",
                "description": "location",
                "maxLength": 100,
                "title": "Location",
                "type": "string",
            },
        },
        "required": ["user"],
        "title": "ProfileSchema",
        "type": "object",
    }

    class ProfileWithUserSchema(ModelSchema):
        user: UserSchema
        model_config = ConfigDict(model=Profile)

    assert ProfileWithUserSchema.model_json_schema() == {
        "$defs": {
            "UserSchema": {
                "description": "A user of the application.",
                "properties": {
                    "profile": {
                        "anyOf": [
                            {
                                "type": "integer",
                            },
                            {
                                "type": "null",
                            },
                        ],
                        "default": None,
                        "description": "id",
                        "title": "Profile",
                    },
                    "id": {
                        "anyOf": [{"type": "integer"}, {"type": "null"}],
                        "default": None,
                        "description": "id",
                        "title": "Id",
                    },
                    "first_name": {
                        "description": "first_name",
                        "maxLength": 50,
                        "title": "First Name",
                        "type": "string",
                    },
                    "last_name": {
                        "anyOf": [
                            {"maxLength": 50, "type": "string"},
                            {"type": "null"},
                        ],
                        "default": None,
                        "description": "last_name",
                        "title": "Last Name",
                    },
                    "email": {
                        "description": "email",
                        "maxLength": 254,
                        "title": "Email",
                        "type": "string",
                    },
                    "created_at": {
                        "description": "created_at",
                        "format": "date-time",
                        "title": "Created At",
                        "type": "string",
                    },
                    "updated_at": {
                        "description": "updated_at",
                        "format": "date-time",
                        "title": "Updated At",
                        "type": "string",
                    },
                },
                "required": ["first_name", "email", "created_at", "updated_at"],
                "title": "UserSchema",
                "type": "object",
            }
        },
        "description": "A user's profile.",
        "properties": {
            "id": {
                "anyOf": [{"type": "integer"}, {"type": "null"}],
                "default": None,
                "description": "id",
                "title": "Id",
            },
            "user": {"$ref": "#/$defs/UserSchema"},
            "website": {
                "default": "",
                "description": "website",
                "maxLength": 200,
                "title": "Website",
                "type": "string",
            },
            "location": {
                "default": "",
                "description": "location",
                "maxLength": 100,
                "title": "Location",
                "type": "string",
            },
        },
        "required": ["user"],
        "title": "ProfileWithUserSchema",
        "type": "object",
    }


@pytest.mark.django_db
def test_one_to_one_reverse():
    """
    Test reverse one-to-one relationships.
    """

    class ProfileSchema(ModelSchema):
        model_config = ConfigDict(model=Profile)

    class UserSchema(ModelSchema):
        model_config = ConfigDict(model=User)

    assert ProfileSchema.model_json_schema() == {
        "description": "A user's profile.",
        "properties": {
            "id": {
                "anyOf": [{"type": "integer"}, {"type": "null"}],
                "default": None,
                "description": "id",
                "title": "Id",
            },
            "user": {"description": "id", "title": "User", "type": "integer"},
            "website": {
                "default": "",
                "description": "website",
                "maxLength": 200,
                "title": "Website",
                "type": "string",
            },
            "location": {
                "default": "",
                "description": "location",
                "maxLength": 100,
                "title": "Location",
                "type": "string",
            },
        },
        "required": ["user"],
        "title": "ProfileSchema",
        "type": "object",
    }

    class UserWithProfileSchema(ModelSchema):
        profile: ProfileSchema
        model_config = ConfigDict(model=User)

    assert UserWithProfileSchema.model_json_schema() == {
        "$defs": {
            "ProfileSchema": {
                "description": "A user's profile.",
                "properties": {
                    "id": {
                        "anyOf": [{"type": "integer"}, {"type": "null"}],
                        "default": None,
                        "description": "id",
                        "title": "Id",
                    },
                    "user": {"description": "id", "title": "User", "type": "integer"},
                    "website": {
                        "default": "",
                        "description": "website",
                        "maxLength": 200,
                        "title": "Website",
                        "type": "string",
                    },
                    "location": {
                        "default": "",
                        "description": "location",
                        "maxLength": 100,
                        "title": "Location",
                        "type": "string",
                    },
                },
                "required": ["user"],
                "title": "ProfileSchema",
                "type": "object",
            }
        },
        "description": "A user of the application.",
        "properties": {
            "profile": {"$ref": "#/$defs/ProfileSchema"},
            "id": {
                "anyOf": [{"type": "integer"}, {"type": "null"}],
                "default": None,
                "description": "id",
                "title": "Id",
            },
            "first_name": {
                "description": "first_name",
                "maxLength": 50,
                "title": "First Name",
                "type": "string",
            },
            "last_name": {
                "anyOf": [{"maxLength": 50, "type": "string"}, {"type": "null"}],
                "default": None,
                "description": "last_name",
                "title": "Last Name",
            },
            "email": {
                "description": "email",
                "maxLength": 254,
                "title": "Email",
                "type": "string",
            },
            "created_at": {
                "description": "created_at",
                "format": "date-time",
                "title": "Created At",
                "type": "string",
            },
            "updated_at": {
                "description": "updated_at",
                "format": "date-time",
                "title": "Updated At",
                "type": "string",
            },
        },
        "required": ["profile", "first_name", "email", "created_at", "updated_at"],
        "title": "UserWithProfileSchema",
        "type": "object",
    }


@pytest.mark.django_db
def test_generic_relation():
    """
    Test generic foreign-key relationships.
    """

    class TaggedSchema(ModelSchema):
        model_config = ConfigDict(model=Tagged)

    assert TaggedSchema.model_json_schema() == {
        "description": "Tagged(id, slug, content_type, object_id)",
        "properties": {
            "id": {
                "anyOf": [{"type": "integer"}, {"type": "null"}],
                "default": None,
                "description": "id",
                "title": "Id",
            },
            "slug": {
                "description": "slug",
                "maxLength": 50,
                "title": "Slug",
                "type": "string",
            },
            "content_type": {
                "description": "id",
                "title": "Content Type",
                "type": "integer",
            },
            "object_id": {
                "description": "object_id",
                "title": "Object Id",
                "type": "integer",
            },
            "content_object": {
                "description": "content_object",
                "title": "Content Object",
                "type": "integer",
            },
        },
        "required": ["slug", "content_type", "object_id", "content_object"],
        "title": "TaggedSchema",
        "type": "object",
    }

    class BookmarkSchema(ModelSchema):
        # FIXME: I added this because for some reason in 2.2 the GenericRelation field
        # ends up required, but in 3 it does not.
        tags: List[Dict[str, int]] = None
        model_config = ConfigDict(model=Bookmark)

    assert BookmarkSchema.model_json_schema() == {
        "description": "Bookmark(id, url)",
        "properties": {
            "id": {
                "anyOf": [{"type": "integer"}, {"type": "null"}],
                "default": None,
                "description": "id",
                "title": "Id",
            },
            "url": {
                "description": "url",
                "maxLength": 200,
                "title": "Url",
                "type": "string",
            },
            "tags": {
                "default": None,
                "items": {
                    "additionalProperties": {"type": "integer"},
                    "type": "object",
                },
                "title": "Tags",
                "type": "array",
            },
        },
        "required": ["url"],
        "title": "BookmarkSchema",
        "type": "object",
    }

    class BookmarkWithTaggedSchema(ModelSchema):

        tags: List[TaggedSchema]
        model_config = ConfigDict(model=Bookmark)

    assert BookmarkWithTaggedSchema.model_json_schema() == {
        "$defs": {
            "TaggedSchema": {
                "description": "Tagged(id, slug, content_type, object_id)",
                "properties": {
                    "id": {
                        "anyOf": [{"type": "integer"}, {"type": "null"}],
                        "default": None,
                        "description": "id",
                        "title": "Id",
                    },
                    "slug": {
                        "description": "slug",
                        "maxLength": 50,
                        "title": "Slug",
                        "type": "string",
                    },
                    "content_type": {
                        "description": "id",
                        "title": "Content Type",
                        "type": "integer",
                    },
                    "object_id": {
                        "description": "object_id",
                        "title": "Object Id",
                        "type": "integer",
                    },
                    "content_object": {
                        "description": "content_object",
                        "title": "Content Object",
                        "type": "integer",
                    },
                },
                "required": ["slug", "content_type", "object_id", "content_object"],
                "title": "TaggedSchema",
                "type": "object",
            }
        },
        "description": "Bookmark(id, url)",
        "properties": {
            "id": {
                "anyOf": [{"type": "integer"}, {"type": "null"}],
                "default": None,
                "description": "id",
                "title": "Id",
            },
            "url": {
                "description": "url",
                "maxLength": 200,
                "title": "Url",
                "type": "string",
            },
            "tags": {
                "items": {"$ref": "#/$defs/TaggedSchema"},
                "title": "Tags",
                "type": "array",
            },
        },
        "required": ["url", "tags"],
        "title": "BookmarkWithTaggedSchema",
        "type": "object",
    }

    class ItemSchema(ModelSchema):

        tags: List[TaggedSchema]
        model_config = ConfigDict(model=Item)

    # Test without defining a GenericRelation on the model
    assert ItemSchema.model_json_schema() == {
        "$defs": {
            "TaggedSchema": {
                "description": "Tagged(id, slug, content_type, object_id)",
                "properties": {
                    "id": {
                        "anyOf": [{"type": "integer"}, {"type": "null"}],
                        "default": None,
                        "description": "id",
                        "title": "Id",
                    },
                    "slug": {
                        "description": "slug",
                        "maxLength": 50,
                        "title": "Slug",
                        "type": "string",
                    },
                    "content_type": {
                        "description": "id",
                        "title": "Content Type",
                        "type": "integer",
                    },
                    "object_id": {
                        "description": "object_id",
                        "title": "Object Id",
                        "type": "integer",
                    },
                    "content_object": {
                        "description": "content_object",
                        "title": "Content Object",
                        "type": "integer",
                    },
                },
                "required": ["slug", "content_type", "object_id", "content_object"],
                "title": "TaggedSchema",
                "type": "object",
            }
        },
        "description": "Item(id, name, item_list)",
        "properties": {
            "id": {
                "anyOf": [{"type": "integer"}, {"type": "null"}],
                "default": None,
                "description": "id",
                "title": "Id",
            },
            "name": {
                "description": "name",
                "maxLength": 100,
                "title": "Name",
                "type": "string",
            },
            "item_list": {"description": "id", "title": "Item List", "type": "integer"},
            "tags": {
                "items": {"$ref": "#/$defs/TaggedSchema"},
                "title": "Tags",
                "type": "array",
            },
        },
        "required": ["name", "item_list", "tags"],
        "title": "ItemSchema",
        "type": "object",
    }


@pytest.mark.django_db
def test_m2m_reverse():
    class ExpertSchema(ModelSchema):
        model_config = ConfigDict(model=Expert)

    class CaseSchema(ModelSchema):
        model_config = ConfigDict(model=Case)

    assert ExpertSchema.model_json_schema() == {
        "description": "Expert(id, name)",
        "properties": {
            "id": {
                "anyOf": [{"type": "integer"}, {"type": "null"}],
                "default": None,
                "description": "id",
                "title": "Id",
            },
            "name": {
                "description": "name",
                "maxLength": 128,
                "title": "Name",
                "type": "string",
            },
            "cases": {
                "description": "id",
                "items": {
                    "additionalProperties": {"type": "integer"},
                    "type": "object",
                },
                "title": "Cases",
                "type": "array",
            },
        },
        "required": ["name", "cases"],
        "title": "ExpertSchema",
        "type": "object",
    }

    assert CaseSchema.model_json_schema() == {
        "description": "Case(id, name, details)",
        "properties": {
            "related_experts": {
                "default": None,
                "description": "id",
                "items": {
                    "additionalProperties": {"type": "integer"},
                    "type": "object",
                },
                "title": "Related Experts",
                "type": "array",
            },
            "id": {
                "anyOf": [{"type": "integer"}, {"type": "null"}],
                "default": None,
                "description": "id",
                "title": "Id",
            },
            "name": {
                "description": "name",
                "maxLength": 128,
                "title": "Name",
                "type": "string",
            },
            "details": {"description": "details", "title": "Details", "type": "string"},
        },
        "required": ["name", "details"],
        "title": "CaseSchema",
        "type": "object",
    }
    case = Case.objects.create(name="My Case", details="Some text data.")
    expert = Expert.objects.create(name="My Expert")
    case_schema = CaseSchema.from_django(case)
    expert_schema = ExpertSchema.from_django(expert)
    assert case_schema.dict() == {
        "related_experts": [],
        "id": 1,
        "name": "My Case",
        "details": "Some text data.",
    }
    assert expert_schema.dict() == {"id": 1, "name": "My Expert", "cases": []}

    expert.cases.add(case)
    case_schema = CaseSchema.from_django(case)
    expert_schema = ExpertSchema.from_django(expert)
    assert case_schema.dict() == {
        "related_experts": [{"id": 1}],
        "id": 1,
        "name": "My Case",
        "details": "Some text data.",
    }
    assert expert_schema.dict() == {"id": 1, "name": "My Expert", "cases": [{"id": 1}]}

    class CustomExpertSchema(ModelSchema):
        """Custom schema"""

        name: Optional[str]
        model_config = ConfigDict(model=Expert)

    class CaseSchema(ModelSchema):
        related_experts: List[CustomExpertSchema]
        model_config = ConfigDict(model=Case)

    assert CaseSchema.model_json_schema() == {
        "$defs": {
            "CustomExpertSchema": {
                "description": "Custom schema",
                "properties": {
                    "id": {
                        "anyOf": [{"type": "integer"}, {"type": "null"}],
                        "default": None,
                        "description": "id",
                        "title": "Id",
                    },
                    "name": {
                        "anyOf": [{"type": "string"}, {"type": "null"}],
                        "default": None,
                        "title": "Name",
                    },
                    "cases": {
                        "description": "id",
                        "items": {
                            "additionalProperties": {"type": "integer"},
                            "type": "object",
                        },
                        "title": "Cases",
                        "type": "array",
                    },
                },
                "required": ["cases"],
                "title": "CustomExpertSchema",
                "type": "object",
            }
        },
        "description": "Case(id, name, details)",
        "properties": {
            "related_experts": {
                "items": {"$ref": "#/$defs/CustomExpertSchema"},
                "title": "Related Experts",
                "type": "array",
            },
            "id": {
                "anyOf": [{"type": "integer"}, {"type": "null"}],
                "default": None,
                "description": "id",
                "title": "Id",
            },
            "name": {
                "description": "name",
                "maxLength": 128,
                "title": "Name",
                "type": "string",
            },
            "details": {"description": "details", "title": "Details", "type": "string"},
        },
        "required": ["related_experts", "name", "details"],
        "title": "CaseSchema",
        "type": "object",
    }

    case_schema = CaseSchema.from_django(case)
    assert case_schema.dict() == {
        "related_experts": [{"id": 1, "name": "My Expert", "cases": [{"id": 1}]}],
        "id": 1,
        "name": "My Case",
        "details": "Some text data.",
    }


@pytest.mark.django_db
def test_alias():
    class ProfileSchema(ModelSchema):
        first_name: str = Field(alias="user__first_name")
        model_config = ConfigDict(model=Profile)

    assert ProfileSchema.model_json_schema() == {
        "description": "A user's profile.",
        "properties": {
            "id": {
                "anyOf": [{"type": "integer"}, {"type": "null"}],
                "default": None,
                "description": "id",
                "title": "Id",
            },
            "user": {"description": "id", "title": "User", "type": "integer"},
            "website": {
                "default": "",
                "description": "website",
                "maxLength": 200,
                "title": "Website",
                "type": "string",
            },
            "location": {
                "default": "",
                "description": "location",
                "maxLength": 100,
                "title": "Location",
                "type": "string",
            },
        },
        "required": ["user", "user__first_name"],
        "title": "ProfileSchema",
        "type": "object",
    }

    user = User.objects.create(first_name="Jack")
    profile = Profile.objects.create(
        user=user, website="www.github.com", location="Europe"
    )
    assert ProfileSchema.from_django(profile).dict() == {
        "first_name": "Jack",
        "id": 1,
        "location": "Europe",
        "user": 1,
        "website": "www.github.com",
    }
